@@ -9,7 +9,7 @@
 #
 #  The original author is Jochen Wiedmann.
 #
-#  Copyright (C) 2009,2010 by H.Merijn Brand & Jens Rehsack
+#  Copyright (C) 2009-2013 by H.Merijn Brand & Jens Rehsack
 #  Copyright (C) 2004 by Jeff Zucker
 #  Copyright (C) 1998 by Jochen Wiedmann
 #
@@ -31,11 +31,11 @@ package DBD::File;
 use strict;
 use warnings;
 
-use base qw(DBI::DBD::SqlEngine);
+use base qw( DBI::DBD::SqlEngine );
 use Carp;
-use vars qw(@ISA $VERSION $drh);
+use vars qw( @ISA $VERSION $drh );
 
-$VERSION = "0.41";
+$VERSION = "0.42";
 
 $drh = undef;		# holds driver handle(s) once initialized
 
@@ -47,7 +47,7 @@ sub driver ($;$)
     # We use a hash here to have one singleton per subclass.
     # (Otherwise DBD::CSV and DBD::DBM, for example, would
     # share the same driver object which would cause problems.)
-    # An alternative would be not not cache the $drh here at all
+    # An alternative would be to not cache the $drh here at all
     # and require that subclasses do that. Subclasses should do
     # their own caching, so caching here just provides extra safety.
     $drh->{$class} and return $drh->{$class};
@@ -83,9 +83,9 @@ package DBD::File::dr;
 use strict;
 use warnings;
 
-use vars qw(@ISA $imp_data_size);
+use vars qw( @ISA $imp_data_size );
 
-@DBD::File::dr::ISA           = qw(DBI::DBD::SqlEngine::dr);
+@DBD::File::dr::ISA           = qw( DBI::DBD::SqlEngine::dr );
 $DBD::File::dr::imp_data_size = 0;
 
 sub dsn_quote
@@ -98,7 +98,7 @@ sub dsn_quote
     } # dsn_quote
 
 # XXX rewrite using TableConfig ...
-sub default_table_source { 'DBD::File::TableSource::FileSystem' }
+sub default_table_source { "DBD::File::TableSource::FileSystem" }
 
 sub disconnect_all
 {
@@ -116,23 +116,24 @@ package DBD::File::db;
 use strict;
 use warnings;
 
-use vars qw(@ISA $imp_data_size);
+use vars qw( @ISA $imp_data_size );
 
 use Carp;
 require File::Spec;
 require Cwd;
-use Scalar::Util qw(refaddr); # in CORE since 5.7.3
+use Scalar::Util qw( refaddr ); # in CORE since 5.7.3
 
-@DBD::File::db::ISA           = qw(DBI::DBD::SqlEngine::db);
+@DBD::File::db::ISA           = qw( DBI::DBD::SqlEngine::db );
 $DBD::File::db::imp_data_size = 0;
 
 sub data_sources
 {
     my ($dbh, $attr, @other) = @_;
-    ref($attr) eq 'HASH' or $attr = {};
-    exists($attr->{f_dir}) or $attr->{f_dir} = $dbh->{f_dir};
-    return $dbh->SUPER::data_sources($attr, @other);
-}
+    ref ($attr) eq "HASH" or $attr = {};
+    exists $attr->{f_dir}        or $attr->{f_dir}     = $dbh->{f_dir};
+    exists $attr->{f_dir_search} or $attr->{f_dir_search} = $dbh->{f_dir_search};
+    return $dbh->SUPER::data_sources ($attr, @other);
+    } # data_source
 
 sub set_versions
 {
@@ -149,6 +150,7 @@ sub init_valid_attributes
     $dbh->{f_valid_attrs} = {
 	f_version        => 1, # DBD::File version
 	f_dir            => 1, # base directory
+	f_dir_search     => 1, # extended search directories
 	f_ext            => 1, # file extension
 	f_schema         => 1, # schema name
 	f_lock           => 1, # Table locking mode
@@ -184,32 +186,34 @@ sub init_default_attributes
     if (0 == $phase) {
 	# f_ext should not be initialized
 	# f_map is deprecated (but might return)
-	$dbh->{f_dir}      = Cwd::abs_path (File::Spec->curdir ());
+	$dbh->{f_dir} = Cwd::abs_path (File::Spec->curdir ());
 
 	push @{$dbh->{sql_init_order}{90}}, "f_meta";
 
-	if(0) { # XXX remove block
 	# complete derived attributes, if required
 	(my $drv_class = $dbh->{ImplementorClass}) =~ s/::db$//;
 	my $drv_prefix = DBI->driver_prefix ($drv_class);
-	my $valid_attrs = $drv_prefix . "valid_attrs";
-	my $ro_attrs = $drv_prefix . "readonly_attrs";
+        if (exists $dbh->{$drv_prefix . "meta"} and !$dbh->{sql_engine_in_gofer}) {
+            my $attr = $dbh->{$drv_prefix . "meta"};
+            defined $dbh->{f_valid_attrs}{f_meta}
+		and $dbh->{f_valid_attrs}{f_meta} = 1;
 
-	my @comp_attrs = ();
-
-	foreach my $comp_attr (@comp_attrs) {
-	    my $attr = $drv_prefix . $comp_attr;
-	    defined $dbh->{$valid_attrs} and !defined $dbh->{$valid_attrs}{$attr} and
-		$dbh->{$valid_attrs}{$attr} = 1;
-	    defined $dbh->{$ro_attrs} and !defined $dbh->{$ro_attrs}{$attr} and
-		$dbh->{$ro_attrs}{$attr} = 1;
+            $dbh->{f_meta} = $dbh->{$attr};
 	    }
-	    } # if 0
 	}
 
     return $dbh;
     } # init_default_attributes
 
+sub validate_FETCH_attr
+{
+    my ($dbh, $attrib) = @_;
+
+    $attrib eq "f_meta" and $dbh->{sql_engine_in_gofer} and $attrib = "sql_meta";
+
+    return $dbh->SUPER::validate_FETCH_attr ($attrib);
+    } # validate_FETCH_attr
+
 sub validate_STORE_attr
 {
     my ($dbh, $attrib, $value) = @_;
@@ -226,6 +230,8 @@ sub validate_STORE_attr
 	    carp "'$value' doesn't look like a valid file extension attribute\n";
 	}
 
+    $attrib eq "f_meta" and $dbh->{sql_engine_in_gofer} and $attrib = "sql_meta";
+
     return $dbh->SUPER::validate_STORE_attr ($attrib, $value);
     } # validate_STORE_attr
 
@@ -240,7 +246,7 @@ sub get_f_versions
     eval {
 	$dver = IO::File->VERSION ();
 
-	# when we're still alive here, everthing went ok - no need to check for $@
+	# when we're still alive here, everything went ok - no need to check for $@
 	$dtype .= " ($dver)";
 	};
 
@@ -264,9 +270,9 @@ package DBD::File::st;
 use strict;
 use warnings;
 
-use vars qw(@ISA $imp_data_size);
+use vars qw( @ISA $imp_data_size );
 
-@DBD::File::st::ISA           = qw(DBI::DBD::SqlEngine::st);
+@DBD::File::st::ISA           = qw( DBI::DBD::SqlEngine::st );
 $DBD::File::st::imp_data_size = 0;
 
 my %supported_attrs = (
@@ -309,7 +315,7 @@ sub FETCH
 			    @colnames ];
 
 	    $attr eq "NULLABLE"  and
-		return [ map { ( grep m/^NOT NULL$/ =>
+		return [ map { ( grep { $_ eq "NOT NULL" }
 			    @{ $sth->{f_overall_defs}{$_}{constraints} || [] })
 			       ? 0 : 1 }
 			    @colnames ];
@@ -328,7 +334,7 @@ use warnings;
 
 use IO::Dir;
 
-@DBD::File::TableSource::FileSystem::ISA = 'DBI::DBD::SqlEngine::TableSource';
+@DBD::File::TableSource::FileSystem::ISA = "DBI::DBD::SqlEngine::TableSource";
 
 sub data_sources
 {
@@ -340,22 +346,28 @@ sub data_sources
     my %attrs;
     $attr and %attrs = %$attr;
     delete $attrs{f_dir};
-    my $dsn_quote = $drh->{ImplementorClass}->can("dsn_quote");
+    my $dsn_quote = $drh->{ImplementorClass}->can ("dsn_quote");
     my $dsnextra = join ";", map { $_ . "=" . &{$dsn_quote} ($attrs{$_}) } keys %attrs;
-    my $dirh = IO::Dir->new($dir);
-    unless (defined $dirh) {
-	$drh->set_err ($DBI::stderr, "Cannot open directory $dir: $!");
-	return;
-	}
+    my @dir = ($dir);
+    $attr->{f_dir_search} && ref $attr->{f_dir_search} eq "ARRAY" and
+	push @dir, grep { -d $_ } @{$attr->{f_dir_search}};
+    my @dsns;
+    foreach $dir (@dir) {
+	my $dirh = IO::Dir->new ($dir);
+	unless (defined $dirh) {
+	    $drh->set_err ($DBI::stderr, "Cannot open directory $dir: $!");
+	    return;
+	    }
 
-    my ($file, @dsns, %names, $driver);
-    $driver = $drh->{ImplementorClass} =~ m/^dbd\:\:([^\:]+)\:\:/i ? $1 : "File";
+	my ($file, %names, $driver);
+	$driver = $drh->{ImplementorClass} =~ m/^dbd\:\:([^\:]+)\:\:/i ? $1 : "File";
 
-    while (defined ($file = $dirh->read())) {
-	my $d = File::Spec->catdir ($dir, $file);
-	# allow current dir ... it can be a data_source too
-	$file ne File::Spec->updir () && -d $d and
-	    push @dsns, "DBI:$driver:f_dir=" . &{$dsn_quote} ($d) . ($dsnextra ? ";$dsnextra" : "");
+	while (defined ($file = $dirh->read ())) {
+	    my $d = File::Spec->catdir ($dir, $file);
+	    # allow current dir ... it can be a data_source too
+	    $file ne File::Spec->updir () && -d $d and
+		push @dsns, "DBI:$driver:f_dir=" . &{$dsn_quote} ($d) . ($dsnextra ? ";$dsnextra" : "");
+	    }
 	}
     return @dsns;
     } # data_sources
@@ -364,35 +376,41 @@ sub avail_tables
 {
     my ($self, $dbh) = @_;
 
-    my $dir    = $dbh->{f_dir};
+    my $dir = $dbh->{f_dir};
     defined $dir or return;	# Stream based db's cannot be queried for tables
-    my $dirh = IO::Dir->new($dir);
 
-    unless (defined $dirh) {
-	$dbh->set_err ($DBI::stderr, "Cannot open directory $dir: $!");
-	return;
-	}
-
-    my $class = $dbh->FETCH ("ImplementorClass");
-    $class =~ s/::db$/::Table/;
-    my ($file, %names);
-    my $schema = exists $dbh->{f_schema}
-	? defined $dbh->{f_schema} && $dbh->{f_schema} ne ""
-	    ? $dbh->{f_schema} : undef
-	: eval { getpwuid ((stat $dir)[4]) }; # XXX Win32::pwent
     my %seen;
     my @tables;
-    while (defined ($file = $dirh->read ())) {
-	my ($tbl, $meta) = $class->get_table_meta ($dbh, $file, 0, 0) or next; # XXX
-	# $tbl && $meta && -f $meta->{f_fqfn} or next;
-	$seen{defined $schema ? $schema : "\0"}{$tbl}++ or
-	    push @tables, [ undef, $schema, $tbl, "TABLE", "FILE" ];
+    my @dir = ($dir);
+    $dbh->{f_dir_search} && ref $dbh->{f_dir_search} eq "ARRAY" and
+	push @dir, grep { -d $_ } @{$dbh->{f_dir_search}};
+    foreach $dir (@dir) {
+	my $dirh = IO::Dir->new ($dir);
+
+	unless (defined $dirh) {
+	    $dbh->set_err ($DBI::stderr, "Cannot open directory $dir: $!");
+	    return;
+	    }
+
+	my $class = $dbh->FETCH ("ImplementorClass");
+	$class =~ s/::db$/::Table/;
+	my ($file, %names);
+	my $schema = exists $dbh->{f_schema}
+	    ? defined $dbh->{f_schema} && $dbh->{f_schema} ne ""
+		? $dbh->{f_schema} : undef
+	    : eval { getpwuid ((stat $dir)[4]) }; # XXX Win32::pwent
+	while (defined ($file = $dirh->read ())) {
+	    my ($tbl, $meta) = $class->get_table_meta ($dbh, $file, 0, 0) or next; # XXX
+	    # $tbl && $meta && -f $meta->{f_fqfn} or next;
+	    $seen{defined $schema ? $schema : "\0"}{$dir}{$tbl}++ or
+		push @tables, [ undef, $schema, $tbl, "TABLE", "FILE" ];
+	    }
+	$dirh->close () or
+	    $dbh->set_err ($DBI::stderr, "Cannot close directory $dir: $!");
 	}
-    $dirh->close() or
-	$dbh->set_err ($DBI::stderr, "Cannot close directory $dir: $!");
 
     return @tables;
-    }
+    } # avail_tables
 
 # ====== DataSource ============================================================
 
@@ -401,21 +419,21 @@ package DBD::File::DataSource::Stream;
 use strict;
 use warnings;
 
-@DBD::File::DataSource::Stream::ISA = 'DBI::DBD::SqlEngine::DataSource';
+use Carp;
+
+@DBD::File::DataSource::Stream::ISA = "DBI::DBD::SqlEngine::DataSource";
 
 # We may have a working flock () built-in but that doesn't mean that locking
 # will work on NFS (flock () may hang hard)
 my $locking = eval {
     my $fh;
-    my $nulldevice = File::Spec->devnull();
-    open( $fh, ">", $nulldevice ) or die "Can't open $nulldevice: $!";
+    my $nulldevice = File::Spec->devnull ();
+    open $fh, ">", $nulldevice or croak "Can't open $nulldevice: $!";
     flock $fh, 0;
     close $fh;
-    1
+    1;
     };
 
-use Carp;
-
 sub complete_table_name
 {
     my ($self, $meta, $file, $respect_case) = @_;
@@ -440,7 +458,7 @@ sub complete_table_name
 sub apply_encoding
 {
     my ($self, $meta, $fn) = @_;
-    defined($fn) or $fn = "file handle " . fileno($meta->{fh});
+    defined $fn or $fn = "file handle " . fileno ($meta->{fh});
     if (my $enc = $meta->{f_encoding}) {
 	binmode $meta->{fh}, ":encoding($enc)" or
 	    croak "Failed to set encoding layer '$enc' on $fn: $!";
@@ -455,26 +473,19 @@ sub open_data
     my ($self, $meta, $attrs, $flags) = @_;
 
     $flags->{dropMode} and croak "Can't drop a table in stream";
-    my $fn = "file handle " . fileno($meta->{f_file});
+    my $fn = "file handle " . fileno ($meta->{f_file});
 
     if ($flags->{createMode} || $flags->{lockMode}) {
-	$meta->{fh} = IO::Handle->new_from_fd( fileno($meta->{f_file} ), "w+" ) or
+	$meta->{fh} = IO::Handle->new_from_fd (fileno ($meta->{f_file}), "w+") or
 	    croak "Cannot open $fn for writing: $! (" . ($!+0) . ")";
 	}
     else {
-	$meta->{fh} = IO::Handle->new_from_fd( fileno($meta->{f_file} ), "r" ) or
+	$meta->{fh} = IO::Handle->new_from_fd (fileno ($meta->{f_file}), "r") or
 	    croak "Cannot open $fn for reading: $! (" . ($!+0) . ")";
 	}
 
-    $meta->{fh} = $meta->{f_file};
     if ($meta->{fh}) {
-	if (my $enc = $meta->{f_encoding}) {
-	    binmode $meta->{fh}, ":encoding($enc)" or
-		croak "Failed to set encoding layer '$enc' on $fn: $!";
-	    }
-	else {
-	    binmode $meta->{fh} or croak "Failed to set binary mode on $fn: $!";
-	    }
+	$self->apply_encoding ($meta, $fn);
 	} # have $meta->{$fh}
 
     if ($self->can_flock && $meta->{fh}) {
@@ -499,7 +510,7 @@ package DBD::File::DataSource::File;
 use strict;
 use warnings;
 
-@DBD::File::DataSource::File::ISA = 'DBD::File::DataSource::Stream';
+@DBD::File::DataSource::File::ISA = "DBD::File::DataSource::Stream";
 
 use Carp;
 
@@ -521,7 +532,7 @@ sub complete_table_name
 	}
 
     # (my $tbl = $file) =~ s/$ext$//i;
-    my ($tbl, $basename, $dir, $fn_ext, $user_spec_file);
+    my ($tbl, $basename, $dir, $fn_ext, $user_spec_file, $searchdir);
     if ($file_is_table and defined $meta->{f_file}) {
 	$tbl = $file;
 	($basename, $dir, $fn_ext) = File::Basename::fileparse ($meta->{f_file}, $fn_any_ext_regex);
@@ -530,6 +541,15 @@ sub complete_table_name
 	}
     else {
 	($basename, $dir, undef) = File::Basename::fileparse ($file, $ext);
+	if ($dir eq "./" && ref $meta->{f_dir_search} eq "ARRAY") {
+	    foreach my $d ($meta->{f_dir}, @{$meta->{f_dir_search}}) {
+		my $f = File::Spec->catdir ($d, $file);
+		-f $f or next;
+		$searchdir = Cwd::abs_path ($d);
+		$dir = "";
+		last;
+		}
+	    }
 	$file = $tbl = $basename;
 	$user_spec_file = 0;
 	}
@@ -543,9 +563,11 @@ sub complete_table_name
         $tbl = lc $tbl;
 	}
 
-    my $searchdir = File::Spec->file_name_is_absolute ($dir)
-	? ($dir =~ s|/$||, $dir)
-	: Cwd::abs_path (File::Spec->catdir ($meta->{f_dir}, $dir));
+    unless (defined $searchdir) {
+	$searchdir = File::Spec->file_name_is_absolute ($dir)
+	    ? ($dir =~ s{/$}{}, $dir)
+	    : Cwd::abs_path (File::Spec->catdir ($meta->{f_dir}, $dir));
+	}
     -d $searchdir or
 	croak "-d $searchdir: $!";
 
@@ -560,7 +582,8 @@ sub complete_table_name
 	if ($respect_case) {
 	    $cmpsub = sub {
 		my ($fn, undef, $sfx) = File::Basename::fileparse ($_, $fn_any_ext_regex);
-		$sfx = '' if $^O eq 'VMS' and $sfx eq '.';  # no extension turns up as a dot
+		$^O eq "VMS" && $sfx eq "." and
+		    $sfx = ""; # no extension turns up as a dot
 		$fn eq $basename and
 		    return (lc $sfx eq lc $ext or !$req && !$sfx);
 		return 0;
@@ -569,7 +592,8 @@ sub complete_table_name
 	else {
 	    $cmpsub = sub {
 		my ($fn, undef, $sfx) = File::Basename::fileparse ($_, $fn_any_ext_regex);
-		$sfx = '' if $^O eq 'VMS' and $sfx eq '.';  # no extension turns up as a dot
+		$^O eq "VMS" && $sfx eq "." and
+		    $sfx = "";  # no extension turns up as a dot
 		lc $fn eq lc $basename and
 		    return (lc $sfx eq lc $ext or !$req && !$sfx);
 		return 0;
@@ -577,10 +601,11 @@ sub complete_table_name
 	    }
 
 	my @f;
-	{
-	    my $dh = IO::Dir->new ($searchdir) or croak "Can't open '$searchdir': $!";
-	    @f = sort { length $b <=> length $a } grep { &$cmpsub ($_) } $dh->read();
-	    $dh->close() or croak "Can't close '$searchdir': $!";
+	{   my $dh = IO::Dir->new ($searchdir) or croak "Can't open '$searchdir': $!";
+	    @f = sort { length $b <=> length $a }
+		 grep { &$cmpsub ($_) }
+		 $dh->read ();
+	    $dh->close () or croak "Can't close '$searchdir': $!";
 	    }
 	@f > 0 && @f <= 2 and $file = $f[0];
 	!$respect_case && $meta->{sql_identifier_case} == 4 and # XXX SQL_IC_MIXED
@@ -607,7 +632,6 @@ sub complete_table_name
     return $tbl;
     } # complete_table_name
 
-
 sub open_data
 {
     my ($self, $meta, $attrs, $flags) = @_;
@@ -635,7 +659,7 @@ sub open_data
 	    $fh->seek (0, 0) or
 		croak "Error while seeking back: $!";
 
-	    $self->apply_encoding($meta);
+	    $self->apply_encoding ($meta);
 	    }
 	}
     if ($meta->{f_fqln}) {
@@ -668,7 +692,7 @@ sub open_data
 	    }
 	# $lm = 0 is forced no locking at all
 	}
-    }
+    } # open_data
 
 # ====== SQL::STATEMENT ========================================================
 
@@ -697,45 +721,43 @@ require Scalar::Util;
 
 # ====== UTILITIES ============================================================
 
-if ( eval { require Params::Util; } )
-{
-    Params::Util->import("_HANDLE");
-}
-else
-{
+if (eval { require Params::Util; }) {
+    Params::Util->import ("_HANDLE");
+    }
+else {
     # taken but modified from Params::Util ...
     *_HANDLE = sub {
 	# It has to be defined, of course
 	defined $_[0] or return;
 
 	# Normal globs are considered to be file handles
-	ref $_[0] eq 'GLOB' and return $_[0];
+	ref $_[0] eq "GLOB" and return $_[0];
 
 	# Check for a normal tied filehandle
-	# Side Note: 5.5.4's tied() and can() doesn't like getting undef
-	tied($_[0]) and tied($_[0])->can('TIEHANDLE') and return $_[0];
+	# Side Note: 5.5.4's tied () and can () doesn't like getting undef
+	tied ($_[0]) and tied ($_[0])->can ("TIEHANDLE") and return $_[0];
 
 	# There are no other non-object handles that we support
-	Scalar::Util::blessed($_[0]) or return;
+	Scalar::Util::blessed ($_[0]) or return;
 
 	# Check for a common base classes for conventional IO::Handle object
-	$_[0]->isa('IO::Handle') and return $_[0];
+	$_[0]->isa ("IO::Handle")  and return $_[0];
 
 	# Check for tied file handles using Tie::Handle
-	$_[0]->isa('Tie::Handle') and return $_[0];
+	$_[0]->isa ("Tie::Handle") and return $_[0];
 
 	# IO::Scalar is not a proper seekable, but it is valid is a
 	# regular file handle
-	$_[0]->isa('IO::Scalar') and return $_[0];
+	$_[0]->isa ("IO::Scalar")  and return $_[0];
 
 	# Yet another special case for IO::String, which refuses (for now
 	# anyway) to become a subclass of IO::Handle.
-	$_[0]->isa('IO::String') and return $_[0];
+	$_[0]->isa ("IO::String")  and return $_[0];
 
 	# This is not any sort of object we know about
 	return;
-    };
-}
+	};
+    }
 
 # ====== FLYWEIGHT SUPPORT =====================================================
 
@@ -747,36 +769,37 @@ sub file2table
 {
     my ($self, $meta, $file, $file_is_table, $respect_case) = @_;
 
-    return $meta->{sql_data_source}->complete_table_name($meta, $file, $respect_case, $file_is_table);
+    return $meta->{sql_data_source}->complete_table_name ($meta, $file, $respect_case, $file_is_table);
     } # file2table
 
 sub bootstrap_table_meta
 {
     my ($self, $dbh, $meta, $table, @other) = @_;
 
-    $self->SUPER::bootstrap_table_meta($dbh, $meta, $table, @other);
+    $self->SUPER::bootstrap_table_meta ($dbh, $meta, $table, @other);
 
-    exists  $meta->{f_dir}	or $meta->{f_dir}	= $dbh->{f_dir};
-    defined $meta->{f_ext}	or $meta->{f_ext}	= $dbh->{f_ext};
-    defined $meta->{f_encoding}	or $meta->{f_encoding}	= $dbh->{f_encoding};
-    exists  $meta->{f_lock}	or $meta->{f_lock}	= $dbh->{f_lock};
-    exists  $meta->{f_lockfile}	or $meta->{f_lockfile}	= $dbh->{f_lockfile};
-    defined $meta->{f_schema}	or $meta->{f_schema}	= $dbh->{f_schema};
+    exists  $meta->{f_dir}        or $meta->{f_dir}        = $dbh->{f_dir};
+    exists  $meta->{f_dir_search} or $meta->{f_dir_search} = $dbh->{f_dir_search};
+    defined $meta->{f_ext}        or $meta->{f_ext}        = $dbh->{f_ext};
+    defined $meta->{f_encoding}   or $meta->{f_encoding}   = $dbh->{f_encoding};
+    exists  $meta->{f_lock}       or $meta->{f_lock}       = $dbh->{f_lock};
+    exists  $meta->{f_lockfile}   or $meta->{f_lockfile}   = $dbh->{f_lockfile};
+    defined $meta->{f_schema}     or $meta->{f_schema}     = $dbh->{f_schema};
 
     defined $meta->{f_open_file_needed} or
 	$meta->{f_open_file_needed} = $self->can ("open_file") != DBD::File::Table->can ("open_file");
 
     defined ($meta->{sql_data_source}) or
 	$meta->{sql_data_source} = _HANDLE ($meta->{f_file})
-	                         ? 'DBD::File::DataSource::Stream'
-				 : 'DBD::File::DataSource::File';
+	                         ? "DBD::File::DataSource::Stream"
+				 : "DBD::File::DataSource::File";
     } # bootstrap_table_meta
 
 sub get_table_meta ($$$$;$)
 {
     my ($self, $dbh, $table, $file_is_table, $respect_case) = @_;
 
-    my $meta = $self->SUPER::get_table_meta($dbh, $table, $respect_case, $file_is_table);
+    my $meta = $self->SUPER::get_table_meta ($dbh, $table, $respect_case, $file_is_table);
     $table = $meta->{table_name};
     return unless $table;
 
@@ -784,17 +807,18 @@ sub get_table_meta ($$$$;$)
     } # get_table_meta
 
 my %reset_on_modify = (
-    f_file     => ["f_fqfn", "sql_data_source"],
-    f_dir      => "f_fqfn",
-    f_ext      => "f_fqfn",
-    f_lockfile => "f_fqfn", # forces new file2table call
+    f_file       => [ "f_fqfn", "sql_data_source" ],
+    f_dir        =>   "f_fqfn",
+    f_dir_search => [],
+    f_ext        =>   "f_fqfn",
+    f_lockfile   =>   "f_fqfn", # forces new file2table call
     );
 
-__PACKAGE__->register_reset_on_modify( \%reset_on_modify );
+__PACKAGE__->register_reset_on_modify (\%reset_on_modify);
 
 my %compat_map = map { $_ => "f_$_" } qw( file ext lock lockfile );
 
-__PACKAGE__->register_compat_map( \%compat_map );
+__PACKAGE__->register_compat_map (\%compat_map);
 
 # ====== DBD::File <= 0.40 compat stuff ========================================
 
@@ -941,19 +965,15 @@ Not really working, always returns an array ref of ones, except the
 affected table has been created in this session.  Valid after
 C<< $sth->execute >>; undef for non-select statements.
 
-=head3 The following DBI attributes and methods are not supported:
+=head3 Unsupported DBI attributes and methods
 
-=over 4
+=head4 bind_param_inout
 
-=item bind_param_inout
+=head4 CursorName
 
-=item CursorName
+=head4 LongReadLen
 
-=item LongReadLen
-
-=item LongTruncOk
-
-=back
+=head4 LongTruncOk
 
 =head3 DBD::File specific attributes
 
@@ -970,8 +990,17 @@ When the value for C<f_dir> is a relative path, it is converted into
 the appropriate absolute path name (based on the current working
 directory) when the dbh attribute is set.
 
+  f_dir => "/data/foo/csv",
+
 See L<KNOWN BUGS AND LIMITATIONS>.
 
+=head4 f_dir_search
+
+This optional attribute can be set to pass a list of folders to also
+find existing tables. It will B<not> be used to create new files.
+
+  f_dir_search => [ "/data/bar/csv", "/dump/blargh/data" ],
+
 =head4 f_ext
 
 This attribute is used for setting the file extension. The format is:
@@ -981,6 +1010,8 @@ This attribute is used for setting the file extension. The format is:
 where the /flag is optional and the extension is case-insensitive.
 C<f_ext> allows you to specify an extension which:
 
+  f_ext => ".csv/r",
+
 =over
 
 =item *
@@ -1068,13 +1099,13 @@ the C<f_lockfile> attribute:
 
   $dbh = DBI->connect ("dbi:DBM:f_lockfile=.foo");
   $dbh->{f_lockfile} = ".foo";
-  $dbh->{f_meta}{qux}{f_lockfile} = ".foo";
+  $dbh->{dbm_tables}{qux}{f_lockfile} = ".foo";
 
 If you wish to disable locking, set the C<f_lockfile> to C<0>.
 
   $dbh = DBI->connect ("dbi:DBM:f_lockfile=0");
   $dbh->{f_lockfile} = 0;
-  $dbh->{f_meta}{qux}{f_lockfile} = 0;
+  $dbh->{dbm_tables}{qux}{f_lockfile} = 0;
 
 =head4 f_encoding
 
@@ -1083,17 +1114,17 @@ This is implemented using C<< binmode $fh, ":encoding(<f_encoding>)" >>.
 
 =head4 f_meta
 
-Private data area which contains information about the tables this
-module handles. Table meta data might not be available until the
-table has been accessed for the first time e.g., by issuing a select
-on it however it is possible to pre-initialize attributes for each table
-you use.
+Private data area aliasing L<DBI::DBD::SqlEngine/sql_meta> which
+contains information about the tables this module handles. Table meta
+data might not be available until the table has been accessed for the
+first time e.g., by issuing a select on it however it is possible to
+pre-initialize attributes for each table you use.
 
 DBD::File recognizes the (public) attributes C<f_ext>, C<f_dir>,
 C<f_file>, C<f_encoding>, C<f_lock>, C<f_lockfile>, C<f_schema>,
-C<col_names>, C<table_name> and C<sql_identifier_case>. Be very careful
-when modifying attributes you do not know, the consequence might be a
-destroyed or corrupted table.
+in addition to the attributes L<DBI::DBD::SqlEngine/sql_meta> already
+supports. Be very careful when modifying attributes you do not know,
+the consequence might be a destroyed or corrupted table.
 
 C<f_file> is an attribute applicable to table meta data only and you
 will not find a corresponding attribute in the dbh. Whilst it may be
@@ -1117,7 +1148,37 @@ write access through another attribute. Well known accessors are
 C<csv_tables> for L<DBD::CSV>, C<ad_tables> for L<DBD::AnyData> and
 C<dbm_tables> for L<DBD::DBM>.
 
-=head3 Internally private attributes to deal with SQL backends:
+=head3 New opportunities for attributes from DBI::DBD::SqlEngine
+
+=head4 sql_table_source
+
+C<< $dbh->{sql_table_source} >> can be set to
+I<DBD::File::TableSource::FileSystem> (and is the default setting
+of DBD::File). This provides usual behaviour of previous DBD::File
+releases on
+
+  @ary = DBI->data_sources ($driver);
+  @ary = DBI->data_sources ($driver, \%attr);
+  
+  @ary = $dbh->data_sources ();
+  @ary = $dbh->data_sources (\%attr);
+
+  @names = $dbh->tables ($catalog, $schema, $table, $type);
+  
+  $sth = $dbh->table_info ($catalog, $schema, $table, $type);
+  $sth = $dbh->table_info ($catalog, $schema, $table, $type, \%attr);
+
+  $dbh->func ("list_tables");
+
+=head4 sql_data_source
+
+C<< $dbh->{sql_data_source} >> can be set to either
+I<DBD::File::DataSource::File>, which is default and provides the
+well known behavior of DBD::File releases prior to 0.41, or
+I<DBD::File::DataSource::Stream>, which reuses already opened
+file-handle for operations.
+
+=head3 Internally private attributes to deal with SQL backends
 
 Do not modify any of these private attributes unless you understand
 the implications of doing so. The behavior of DBD::File and derived
@@ -1160,17 +1221,6 @@ If you want to read the subdirectories of another directory, use
     my ($drh)  = DBI->install_driver ("CSV");
     my (@list) = $drh->data_sources (f_dir => "/usr/local/csv_data");
 
-=head4 list_tables
-
-This method returns a list of file names inside $dbh->{f_dir}.
-Example:
-
-    my ($dbh)  = DBI->connect ("dbi:CSV:f_dir=/usr/local/csv_data");
-    my (@list) = $dbh->func ("list_tables");
-
-Note that the list includes all files contained in the directory, even
-those that have non-valid table names, from the view of SQL.
-
 =head3 Additional methods
 
 The following methods are only available via their documented name when
@@ -1182,11 +1232,11 @@ by replacing the C<f_> in the method name with the driver prefix.
 
 Signature:
 
-    sub f_versions (;$)
-    {
-	my ($table_name) = @_;
-	$table_name ||= ".";
-	...
+  sub f_versions (;$)
+  {
+    my ($table_name) = @_;
+    $table_name ||= ".";
+    ...
     }
 
 Returns the versions of the driver, including the DBI version, the Perl
@@ -1194,13 +1244,14 @@ version, DBI::PurePerl version (if DBI::PurePerl is active) and the version
 of the SQL engine in use.
 
     my $dbh = DBI->connect ("dbi:File:");
-    my $f_versions = $dbh->f_versions ();
+    my $f_versions = $dbh->func ("f_versions");
     print "$f_versions\n";
     __END__
-    # DBD::File        0.39 using SQL::Statement 1.28
-    # DBI              1.612
-    # OS               netbsd (5.99.24)
-    # Perl             5.010001 (x86_64-netbsd-thread-multi)
+    # DBD::File              0.41 using IO::File (1.16)
+    #   DBI::DBD::SqlEngine  0.05 using SQL::Statement 1.406
+    # DBI                    1.623
+    # OS                     darwin (12.2.1)
+    # Perl                   5.017006 (darwin-thread-multi-ld-2level)
 
 Called in list context, f_versions will return an array containing each
 line as single entry.
@@ -1209,65 +1260,6 @@ Some drivers might use the optional (table name) argument and modify
 version information related to the table (e.g. DBD::DBM provides storage
 backend information for the requested table, when it has a table name).
 
-=head4 f_get_meta
-
-Signature:
-
-    sub f_get_meta ($$)
-    {
-	my ($table_name, $attrib) = @_;
-	...
-    }
-
-Returns the value of a meta attribute set for a specific table, if any.
-See L<f_meta> for the possible attributes.
-
-A table name of C<"."> (single dot) is interpreted as the default table.
-This will retrieve the appropriate attribute globally from the dbh.
-This has the same restrictions as C<< $dbh->{$attrib} >>.
-
-=head4 f_set_meta
-
-Signature:
-
-    sub f_set_meta ($$$)
-    {
-	my ($table_name, $attrib, $value) = @_;
-	...
-    }
-
-Sets the value of a meta attribute set for a specific table.
-See L<f_meta> for the possible attributes.
-
-A table name of C<"."> (single dot) is interpreted as the default table
-which will set the specified attribute globally for the dbh.
-This has the same restrictions as C<< $dbh->{$attrib} = $value >>.
-
-=head4 f_clear_meta
-
-Signature:
-
-    sub f_clear_meta ($)
-    {
-	my ($table_name) = @_;
-	...
-    }
-
-Clears the table specific meta information in the private storage of the
-dbh.
-
-=head1 SQL ENGINES
-
-DBD::File currently supports two SQL engines: L<SQL::Statement|SQL::Statement>
-and L<DBI::SQL::Nano::Statement_|DBI::SQL::Nano>. DBI::SQL::Nano supports a
-I<very> limited subset of SQL statements, but it might be faster for some
-very simple tasks. SQL::Statement in contrast supports a much larger subset
-of ANSI SQL.
-
-To use SQL::Statement, you need at least version 1.28 of
-SQL::Statement and the environment variable C<DBI_SQL_NANO> must not
-be set to a true value.
-
 =head1 KNOWN BUGS AND LIMITATIONS
 
 =over 4
@@ -1300,6 +1292,8 @@ evaluated instead of driver globals:
 
 =item f_dir
 
+=item f_dir_search
+
 =item f_lock
 
 =item f_lockfile
@@ -1364,13 +1358,13 @@ tables, but they will not covered by C<table_info>.
 This module is currently maintained by
 
 H.Merijn Brand < h.m.brand at xs4all.nl > and
-Jens Rehsack  < rehsack at googlemail.com >
+Jens Rehsack < rehsack at googlemail.com >
 
 The original author is Jochen Wiedmann.
 
 =head1 COPYRIGHT AND LICENSE
 
- Copyright (C) 2009-2010 by H.Merijn Brand & Jens Rehsack
+ Copyright (C) 2009-2013 by H.Merijn Brand & Jens Rehsack
  Copyright (C) 2004-2009 by Jeff Zucker
  Copyright (C) 1998-2004 by Jochen Wiedmann
 
